package adaptors;

import java.util.EnumSet;
import java.util.Set;

import net.minecraft.tileentity.TileEntity;

import net.minecraftforge.common.MinecraftForge;
import net.minecraftforge.common.util.ForgeDirection;

import cofh.api.energy.IEnergyHandler;
import enetbridge.EnetBridge;
import ic2.api.energy.event.EnergyTileLoadEvent;
import ic2.api.energy.event.EnergyTileUnloadEvent;

public class RfBlockDelegate extends TileEntityDelegate {
	public static TileEntityDelegate create(TileEntity te) {
		if (!isApplicable(te)) return null;

		try {
			return new RfBlockDelegate(te);
		} catch (Exception e) {
			EnetBridge.log.warn("RF delegate init error: {}.", e.getMessage());
			return null;
		}
	}

	public static boolean isApplicable(TileEntity te) {
		return EnetBridge.hasRf() && te instanceof IEnergyHandler;
	}

	private RfBlockDelegate(TileEntity base) {
		super(base);

		this.base = (IEnergyHandler) base;
		updateDirs(false);
	}

	@Override
	public void tick() {
		if (!faulty && updateDirs(true)) {
			// refresh connectivity
			MinecraftForge.EVENT_BUS.post(new EnergyTileUnloadEvent(this));
			updateDirs(false);
			MinecraftForge.EVENT_BUS.post(new EnergyTileLoadEvent(this));
		}
	}

	@Override
	public boolean acceptsEnergyFrom(TileEntity emitter, ForgeDirection direction) {
		return !faulty && sinkDirs.contains(direction);
	}

	@Override
	public boolean emitsEnergyTo(TileEntity receiver, ForgeDirection direction) {
		return !faulty && sourceDirs.contains(direction);
	}

	@Override
	public double getDemandedEnergy() {
		if (faulty) return 0;

		int max = 0;

		for (ForgeDirection dir : sinkDirs) {
			int amount = base.receiveEnergy(dir, Integer.MAX_VALUE, true);
			if (amount > max) max = amount;
		}

		return max * 1. / EnetBridge.getRfPerEu();
	}

	@Override
	public double getOfferedEnergy() {
		if (faulty) return 0;

		int max = 0;

		for (ForgeDirection dir : sourceDirs) {
			int amount = base.extractEnergy(dir, Integer.MAX_VALUE, true);
			if (amount > max) max = amount;
		}

		return max * EnetBridge.getEuPerRf();
	}

	@Override
	public double injectEnergy(ForgeDirection directionFrom, double amount, double voltage) {
		amount *= EnetBridge.getRfPerEu();
		int iAmount = (int) amount;
		if (iAmount <= 0) return amount;

		amount -= iAmount;

		if (directionFrom != ForgeDirection.UNKNOWN) {
			iAmount -= base.receiveEnergy(directionFrom, iAmount, false);
		}

		for (ForgeDirection dir : sinkDirs) {
			if (iAmount <= 0) break;
			iAmount -= base.receiveEnergy(dir, iAmount, false);
		}

		return (amount + iAmount) * 1. / EnetBridge.getRfPerEu();
	}

	@Override
	public void drawEnergy(double amount) {
		for (ForgeDirection dir : sourceDirs) {
			amount -= base.extractEnergy(dir, (int) Math.ceil(amount / EnetBridge.getEuPerRf()), false) * EnetBridge.getEuPerRf();
			if (amount <= 0) break;
		}

		if (amount > 0) {
			EnetBridge.log.warn("Can't draw the full amount from {}, {} left over, disabling the delegate.", base, amount);
			faulty = true;
		}
	}

	private boolean updateDirs(boolean simulate) {
		EnumSet<ForgeDirection> newSinkdirs = EnumSet.noneOf(ForgeDirection.class);
		EnumSet<ForgeDirection> newSourcedirs = EnumSet.noneOf(ForgeDirection.class);

		// probe directions if they are connectable and can source or sink energy
		for (ForgeDirection dir : ForgeDirection.VALID_DIRECTIONS) {
			if (!base.canConnectEnergy(dir)) continue;

			if (EnetBridge.getRfPerEu() > 0) { // sink
				// if base is full we draw 1 RF to allow injecting
				//int undo = base.getEnergyStored(dir) < base.getMaxEnergyStored(dir) ? 0 : extractFromAnySide(dir, 1);

				if (base.receiveEnergy(dir, Integer.MAX_VALUE, true) > 0) {
					newSinkdirs.add(dir);
				}

				// fill base back up if applicable
				/*if (undo > 0 && receiveFromAnySide(dir, undo) != undo) {
					EnetBridge.log.warn("Can't revert receive test for {}, {} missing, disabling the delegate.", base, undo);
					faulty = true;
					return false;
				}*/
			}

			if (EnetBridge.getEuPerRf() > 0) { // source
				// if base is empty we inject 1 RF to allow drawing
				//int undo = base.getEnergyStored(dir) > 0 ? 0 : receiveFromAnySide(dir, 1);

				if (base.extractEnergy(dir, Integer.MAX_VALUE, true) > 0) {
					newSourcedirs.add(dir);
				}

				// empty base again if applicable
				/*if (undo > 0 && extractFromAnySide(dir, undo) != undo) {
					EnetBridge.log.warn("Can't revert extract test for {}, {} missing, disabling the delegate.", base, undo);
					faulty = true;
					return false;
				}*/
			}
		}

		if (sinkDirs.equals(newSinkdirs) && sourceDirs.equals(newSourcedirs)) {
			return false;
		} else {
			if (!simulate) {
				sinkDirs.clear();
				sinkDirs.addAll(newSinkdirs);

				sourceDirs.clear();
				sourceDirs.addAll(newSourcedirs);
			}

			EnetBridge.log.debug("Detected new dirs for {}: {}, {}.", base, sinkDirs, sourceDirs);

			return true;
		}
	}

	private int extractFromAnySide(ForgeDirection preferedSide, int amount) {
		int remAmount = amount - base.extractEnergy(preferedSide, amount, false);

		for (ForgeDirection dir : ForgeDirection.VALID_DIRECTIONS) {
			if (remAmount <= 0) break;
			if (dir == preferedSide) continue;

			remAmount -= base.extractEnergy(dir, remAmount, false);
		}

		return amount - remAmount;
	}

	private int receiveFromAnySide(ForgeDirection preferedSide, int amount) {
		int remAmount = amount - base.receiveEnergy(preferedSide, amount, false);

		for (ForgeDirection dir : ForgeDirection.VALID_DIRECTIONS) {
			if (remAmount <= 0) break;
			if (dir == preferedSide) continue;

			remAmount -= base.receiveEnergy(dir, remAmount, false);
		}

		return amount - remAmount;
	}


	private final IEnergyHandler base;
	protected boolean faulty = false;
	protected final Set<ForgeDirection> sinkDirs = EnumSet.noneOf(ForgeDirection.class);
	protected final Set<ForgeDirection> sourceDirs = EnumSet.noneOf(ForgeDirection.class);
}
